LIBRARY libiscsi
EXPORTS
iscsi_get_fd
iscsi_which_events
iscsi_service
iscsi_queue_length
iscsi_parse_full_url
iscsi_destroy_url
iscsi_parse_portal_url
iscsi_get_error
iscsi_create_context
iscsi_destroy_context
iscsi_set_alias
iscsi_set_targetname
iscsi_get_target_address
iscsi_set_session_type
iscsi_set_header_digest
iscsi_set_initiator_username_pwd
iscsi_is_logged_in
iscsi_connect_async
iscsi_connect_sync
iscsi_full_connect_async
iscsi_full_connect_sync
iscsi_disconnect
iscsi_login_async
iscsi_login_sync
iscsi_logout_async
iscsi_logout_sync
iscsi_discovery_async
iscsi_nop_out_async
iscsi_task_mgmt_async
iscsi_task_mgmt_abort_task_async
iscsi_task_mgmt_abort_task_set_async
iscsi_task_mgmt_lun_reset_async
iscsi_task_mgmt_target_warm_reset_async
iscsi_task_mgmt_target_cold_reset_async
iscsi_set_isid_oui
iscsi_set_isid_en
iscsi_set_isid_random
iscsi_set_isid_reserved
iscsi_scsi_command_async
iscsi_reportluns_task
iscsi_testunitready_task
iscsi_inquiry_task
iscsi_readcapacity10_task
iscsi_readcapacity16_task
iscsi_synchronizecache10_task
iscsi_read6_task
iscsi_read10_task
iscsi_read12_task
iscsi_verify10_task
iscsi_write10_task
iscsi_write12_task
iscsi_writesame10_task
iscsi_writesame16_task
iscsi_modesense6_task
iscsi_scsi_command_sync
iscsi_reportluns_sync
iscsi_testunitready_sync
iscsi_inquiry_sync
iscsi_readcapacity10_sync
iscsi_readcapacity16_sync
iscsi_synchronizecache10_sync
iscsi_read6_sync
iscsi_read10_sync
iscsi_read12_sync
iscsi_verify10_sync
iscsi_write10_sync
iscsi_write12_sync
iscsi_writesame10_sync
iscsi_writesame16_sync
iscsi_scsi_task_cancel
iscsi_scsi_cancel_all_tasks
poll
scsi_task_add_data_in_buffer
scsi_sense_key_str
scsi_sense_ascq_str
scsi_free_scsi_task
scsi_set_task_private_ptr
scsi_get_task_private_ptr
scsi_cdb_testunitready
scsi_reportluns_cdb
scsi_cdb_readcapacity10
scsi_devtype_to_str
scsi_version_to_str
scsi_inquiry_pagecode_to_str
scsi_cdb_inquiry
scsi_protocol_identifier_to_str
scsi_codeset_to_str
scsi_association_to_str
scsi_designator_type_to_str
scsi_cdb_modesense6
scsi_datain_getfullsize
scsi_datain_unmarshall
scsi_cdb_read6
scsi_cdb_read10
scsi_cdb_verify10
scsi_cdb_write10
scsi_cdb_writesame10
scsi_cdb_writesame16
scsi_cdb_synchronizecache10
scsi_cdb_readcapacity16
scsi_cdb_serviceactionin16
scsi_cdb_unmap

